"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AssetManager = void 0;
const tslib_1 = require("tslib");
const fs_1 = tslib_1.__importDefault(require("fs"));
const path_1 = tslib_1.__importDefault(require("path"));
const BundleFile_1 = require("./BundleFile");
const EndianReader_1 = require("./EndianReader");
const ObjectReader_1 = require("./ObjectReader");
const Objects_1 = require("./Objects");
const NamedObject_1 = require("./Objects/NamedObject");
const SerializedFile_1 = require("./SerializedFile");
const fast_glob_1 = tslib_1.__importDefault(require("fast-glob"));
const Sprite_1 = require("./Objects/Sprite");
class AssetManager {
    constructor() {
        this.assetFiles = new Map();
        this.resourceFiles = new Map();
        this.resourceFileReaders = new Map();
        this.namedObjects = new Map();
    }
    async loadFiles(rootdir, assetDirGlobs, assetGlobs) {
        const assetDirectories = await fast_glob_1.default(assetDirGlobs, { cwd: rootdir, onlyDirectories: true });
        let bundleDirs = [];
        for (const assetDir of assetDirectories) {
            const dirs = await fast_glob_1.default(assetGlobs, { cwd: path_1.default.join(rootdir, assetDir), onlyDirectories: true });
            bundleDirs.push(...dirs.map((dir) => path_1.default.join(rootdir, assetDir, dir)));
        }
        const jobs = [];
        for (const dir of bundleDirs) {
            const [bundlePrefix] = path_1.default.basename(dir).split('_');
            const bundleFiles = await fast_glob_1.default(['**/__data', '**/*.unity3d'], { cwd: dir });
            for (const file of bundleFiles) {
                jobs.push(this.loadFile(path_1.default.join(dir, file), undefined, bundlePrefix));
            }
        }
        const results = await Promise.allSettled(jobs);
        const failures = results.filter((result) => result.status === 'rejected');
        if (failures.length) {
            console.warn(`Failed to load ${failures.length} bundles.`);
        }
    }
    async loadFile(filePath, buffer, bundlePrefix) {
        if (!buffer) {
            buffer = await fs_1.default.promises.readFile(filePath);
        }
        const reader = EndianReader_1.EndianReader.fromBuffer(buffer);
        const bundle = await BundleFile_1.BundleFile.create(reader);
        for (const file of bundle.files) {
            const reader = EndianReader_1.EndianReader.fromBuffer(file.buffer);
            if (SerializedFile_1.isSerializedFile(reader)) {
                const containerName = path_1.default.join(filePath, file.name);
                const assetFile = new SerializedFile_1.SerializedFile(this, file.name, reader, bundlePrefix);
                assetFile.setVersion(bundle.header.unityVersion[1]);
                this.assetFiles.set(containerName, assetFile);
            }
            else {
                this.resourceFileReaders.set(file.name, reader);
            }
        }
        this.resourceFiles.set(filePath, bundle);
    }
    loadAssets() {
        for (const [_containerKey, assetFile] of this.assetFiles.entries()) {
            for (const objectInfo of assetFile.objectInfos) {
                const reader = new ObjectReader_1.ObjectReader(assetFile.reader, assetFile, objectInfo);
                const object = Objects_1.createObject(objectInfo.classId, reader);
                assetFile.objects.set(objectInfo.pathId, object);
                if (NamedObject_1.isNamedObject(object)) {
                    let name = object.name;
                    // sprites may have name collisions, and won't be referenced by other resources
                    if (Sprite_1.isSprite(object)) {
                        name = [assetFile.bundlePrefix, object.name].filter(Boolean).join('/');
                    }
                    this.namedObjects.set(name, object);
                }
            }
        }
    }
}
exports.AssetManager = AssetManager;
//# sourceMappingURL=AssetManager.js.map