"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BundleFile = void 0;
const tslib_1 = require("tslib");
const path_1 = tslib_1.__importDefault(require("path"));
const smart_buffer_1 = require("smart-buffer");
const EndianReader_1 = require("./EndianReader");
const Assert_1 = require("../utils/Assert");
const lz4 = require('lz4');
const lzma = require('lzma/src/lzma_worker').LZMA_WORKER;
class BundleFile {
    constructor(reader) {
        this.reader = reader;
    }
    static async create(reader) {
        const file = new BundleFile(reader);
        await file.init();
        return file;
    }
    async init() {
        this.header = this.readHeader();
        this.blocksInfo = this.readBlocksInfo();
        this.blocksStream = await this.readBlocksStream();
        this.files = this.readFiles();
    }
    readFiles() {
        const files = [];
        for (const node of this.blocksInfo.nodes) {
            const name = path_1.default.basename(node.path);
            const buffer = this.blocksStream.internalBuffer.slice(Number(node.offset), Number(node.offset + node.size));
            files.push({ name, buffer });
        }
        return files;
    }
    async readBlocksStream() {
        const uncompressedSize = this.blocksInfo.blocks.reduce((sum, value) => sum + value.size.uncompressed, 0);
        const blocksStream = smart_buffer_1.SmartBuffer.fromBuffer(Buffer.alloc(uncompressedSize));
        for (const block of this.blocksInfo.blocks) {
            const blockBuffer = this.reader.readBuffer(block.size.compressed);
            switch (block.flags & 0x3f // kStorageBlockCompressionTypeMask
            ) {
                case 1:
                    // const headerBuf = blockBuffer.slice(0, 5);
                    // const dataBuffer = blockBuffer.slice(13);
                    // const bufsize = dataBuffer.length;
                    const result = Buffer.from(lzma.decompress(blockBuffer));
                    Assert_1.assert(result.length === block.size.uncompressed);
                    blocksStream.writeBuffer(result);
                    break;
                case 2:
                case 3: {
                    const decodedSize = lz4.decodeBlock(blockBuffer, blocksStream.internalBuffer.slice(blocksStream.writeOffset));
                    blocksStream.writeOffset += decodedSize;
                    Assert_1.assert(decodedSize === block.size.uncompressed);
                    break;
                }
                default: {
                    this.reader.writeBuffer(blockBuffer);
                    // None
                    break;
                }
            }
        }
        blocksStream.readOffset = 0;
        return blocksStream;
    }
    alignStream(alignment) {
        const pos = this.reader.readOffset;
        const mod = pos % alignment;
        if (mod !== 0) {
            this.reader.readOffset += alignment - mod;
        }
    }
    readBlocksInfo() {
        const stream = this.getBlocksInfoStream();
        const reader = EndianReader_1.EndianReader.fromBuffer(stream);
        reader.readOffset += 16; // uncompressedDataHash
        const blocksInfoCount = reader.readInt32();
        const blocks = [];
        for (let n = 0; n < blocksInfoCount; ++n) {
            blocks.push({
                size: { uncompressed: reader.readUInt32(), compressed: reader.readUInt32() },
                flags: reader.readUInt16(),
            });
        }
        const nodesCount = reader.readInt32();
        const nodes = [];
        for (let n = 0; n < nodesCount; ++n) {
            nodes.push({
                offset: reader.readInt64(),
                size: reader.readInt64(),
                flags: reader.readUInt32(),
                path: reader.readStringNT(),
            });
        }
        return { blocks, nodes };
    }
    getBlocksInfoStream() {
        let blocksInfoCompressedStream;
        let blocksInfoUncompressedStream;
        if (this.header.version >= 7) {
            this.alignStream(16);
        }
        if (this.header.flags & 0x80) {
            // kArchiveBlocksInfoAtTheEnd
            blocksInfoCompressedStream = this.reader.internalBuffer.slice(this.reader.length - this.header.blocksInfoSize.compressed, this.reader.length);
        }
        else {
            // 0x40 kArchiveBlocksAndDirectoryInfoCombined
            blocksInfoCompressedStream = this.reader.readBuffer(this.header.blocksInfoSize.compressed);
        }
        // kArchiveCompressionTypeMask
        switch (this.header.flags & 0x3f) {
            default: {
                // None
                blocksInfoUncompressedStream = blocksInfoCompressedStream;
                break;
            }
            case 1: {
                blocksInfoUncompressedStream = Buffer.from(lzma.decompress(blocksInfoCompressedStream));
                break;
            }
            case 2: // LZ4
            case 3: {
                // LZ4HC
                blocksInfoUncompressedStream = Buffer.alloc(this.header.blocksInfoSize.uncompressed);
                const decodedSize = lz4.decodeBlock(blocksInfoCompressedStream, blocksInfoUncompressedStream);
                Assert_1.assert(decodedSize === this.header.blocksInfoSize.uncompressed);
                break;
            }
        }
        return blocksInfoUncompressedStream;
    }
    readHeader() {
        const signature = this.reader.readStringNT();
        const version = this.reader.readUInt32();
        const unityVersion = [this.reader.readStringNT(), this.reader.readStringNT()];
        switch (signature) {
            case 'UnityWeb':
            case 'UnityRaw':
                if (version != 6) {
                    throw new Error(`Archive type not supported: '${signature}'`);
                }
            case 'UnityFS':
                const size = this.reader.readInt64();
                const blocksInfoSize = { compressed: this.reader.readUInt32(), uncompressed: this.reader.readUInt32() };
                const flags = this.reader.readUInt32();
                if (signature != 'UnityFS') {
                    this.reader.readInt8();
                }
                return { signature, version, unityVersion, blocksInfoSize, size, flags };
        }
        throw new Error(`Internal error reading header`);
    }
}
exports.BundleFile = BundleFile;
//# sourceMappingURL=BundleFile.js.map