"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SerializedFile = exports.isSerializedFile = void 0;
const smart_buffer_1 = require("smart-buffer");
const BuildTarget_1 = require("./BuildTarget");
const BuildType_1 = require("./BuildType");
const CommonString_1 = require("./CommonString");
const VersionTuple_1 = require("./VersionTuple");
function isSerializedFile(reader) {
    const bufferSize = reader.length;
    if (bufferSize < 20) {
        return false;
    }
    reader.readUInt32(); // metadataSize
    let fileSize = reader.readUInt32();
    const version = reader.readUInt32();
    let dataOffset = reader.readUInt32();
    reader.readInt8(); //endianness
    reader.readBuffer(3); // reserved
    if (version >= 22) {
        if (fileSize < 48) {
            return false;
        }
        reader.readUInt32(); // metadataSize
        fileSize = reader.readInt64();
        dataOffset = reader.readInt64();
    }
    reader.readOffset = 0;
    if (BigInt(fileSize) !== BigInt(bufferSize)) {
        return false;
    }
    if (dataOffset > fileSize) {
        return false;
    }
    return true;
}
exports.isSerializedFile = isSerializedFile;
class SerializedFile {
    constructor(assetManager, name, reader, bundlePrefix) {
        this.assetManager = assetManager;
        this.name = name;
        this.bundlePrefix = bundlePrefix;
        this.enableTypeTree = true;
        this.bigIDEnabled = false;
        this.version = VersionTuple_1.makeVersion();
        this.unityVersion = '2.5.0f5';
        this.targetPlatform = BuildTarget_1.BuildTarget.UnknownPlatform; // unknown
        this.objects = new Map();
        this.buildType = BuildType_1.BuildType.None;
        this.reader = reader;
        this.header = this.readHeader();
        this.readMetadata();
        this.types = this.readTypes();
        this.objectInfos = this.readObjects();
        // TODO: Read externals
        // TODO: Read ref types
        // TODO: Read userinformation stub(?)
    }
    setVersion(version) {
        const buildFlag = version.replace(/[\d\.]/g, '');
        if (buildFlag === BuildType_1.BuildType.Alpha || buildFlag === BuildType_1.BuildType.Patch) {
            this.buildType = buildFlag;
        }
        else {
            this.buildType = BuildType_1.BuildType.None;
        }
        this.version = VersionTuple_1.makeVersion(...version
            .replace(/\D/g, '.')
            .split('.')
            .filter(Boolean)
            .map((part) => parseInt(part, 10)));
    }
    readHeader() {
        let metadataSize = this.reader.readUInt32();
        let fileSize = this.reader.readUInt32();
        let version = this.reader.readUInt32();
        let dataOffset = this.reader.readUInt32();
        let endianness;
        if (version >= 9) {
            endianness = this.reader.readUInt8();
            this.reader.readOffset += 3; // reserved[3]
        }
        else {
            this.reader.readOffset = fileSize - metadataSize;
            endianness = this.reader.readUInt8();
        }
        if (version >= 22) {
            metadataSize = this.reader.readUInt32();
            fileSize = this.reader.readInt64();
            dataOffset = this.reader.readInt64();
            this.reader.readInt64(); // unknown
        }
        this.reader.endianness = endianness;
        return { metadataSize, fileSize, version, dataOffset, endianness };
    }
    readMetadata() {
        if (this.header.version >= 7) {
            this.unityVersion = this.reader.readStringNT();
            this.setVersion(this.unityVersion);
        }
        if (this.header.version >= 8) {
            this.targetPlatform = this.reader.readInt32();
        }
        if (this.header.version >= 13) {
            this.enableTypeTree = this.reader.readBoolean();
        }
    }
    readTypes() {
        const typeCount = this.reader.readInt32();
        const types = [];
        for (let n = 0; n < typeCount; ++n) {
            types.push(this.readSerializedType());
        }
        if (this.header.version >= 7 && this.header.version < 14) {
            this.bigIDEnabled = !!this.reader.readInt32();
        }
        return types;
    }
    readSerializedType() {
        const classId = this.reader.readInt32();
        const type = {
            classId,
        };
        if (this.header.version >= 16) {
            type.isStrippedType = this.reader.readBoolean();
        }
        if (this.header.version >= 17) {
            type.scriptTypeIndex = this.reader.readInt16();
        }
        if (this.header.version >= 13) {
            if ((this.header.version > 16 && classId < 0) || (this.header.version >= 16 && classId === 114)) {
                type.scriptId = this.reader.readBuffer(16);
            }
            type.oldTypeHash = this.reader.readBuffer(16);
        }
        if (this.enableTypeTree) {
            if (this.header.version >= 12 || this.header.version === 10) {
                type.nodes = this.typeTreeBlobRead();
            }
            else {
                type.nodes = this.readTypeTree([]);
            }
            if (this.header.version >= 21) {
                type.typeDependencies = this.reader.readArray('readInt32');
            }
        }
        return type;
    }
    typeTreeBlobRead() {
        const numberOfNodes = this.reader.readInt32();
        const stringBufferSize = this.reader.readInt32();
        const nodes = [];
        for (let n = 0; n < numberOfNodes; ++n) {
            const version = this.reader.readUInt16();
            const level = this.reader.readInt8();
            const isArray = this.reader.readBoolean();
            const typeStrOffset = this.reader.readUInt32();
            const nameStrOffset = this.reader.readUInt32();
            const byteSize = this.reader.readInt32();
            const index = this.reader.readInt32();
            const metaFlag = this.reader.readInt32();
            if (this.header.version >= 19) {
                this.reader.readUInt64(); // refTypeHash
            }
            nodes.push({
                level,
                version,
                isArray,
                typeStrOffset,
                nameStrOffset,
                byteSize,
                index,
                metaFlag,
            });
        }
        const stringBuffer = smart_buffer_1.SmartBuffer.fromBuffer(this.reader.readBuffer(stringBufferSize));
        function readString(value) {
            var _a;
            const isOffset = (value & 0x80000000) == 0;
            if (isOffset) {
                stringBuffer.readOffset = value;
                return stringBuffer.readStringNT();
            }
            const offset = value & 0x7fffffff;
            return (_a = CommonString_1.commonStringMap[offset]) !== null && _a !== void 0 ? _a : offset.toString();
        }
        for (const node of nodes) {
            node.type = readString(node.typeStrOffset);
            node.name = readString(node.nameStrOffset);
        }
        return nodes;
    }
    readTypeTree(nodes, level = 0) {
        const type = this.reader.readStringNT();
        const name = this.reader.readStringNT();
        const byteSize = this.reader.readInt32();
        let index;
        if (this.header.version === 2) {
            this.reader.readInt32(); // variableCount
        }
        if (this.header.version !== 3) {
            index = this.reader.readInt32();
        }
        const isArray = !!this.reader.readInt32();
        const version = this.reader.readInt32();
        let metaFlag;
        if (this.header.version !== 3) {
            metaFlag = this.reader.readInt32();
        }
        nodes.push({
            level,
            type,
            name,
            byteSize,
            index,
            isArray,
            version,
            metaFlag,
        });
        const childrenCount = this.reader.readInt32();
        for (let n = 0; n < childrenCount; ++n) {
            this.readTypeTree(nodes, level + 1);
        }
        return nodes;
    }
    readObjects() {
        const objectCount = this.reader.readInt32();
        const objects = [];
        for (let n = 0; n < objectCount; ++n) {
            let pathId;
            if (this.bigIDEnabled) {
                pathId = this.reader.readInt64();
            }
            else if (this.header.version < 14) {
                pathId = this.reader.readInt32();
            }
            else {
                this.reader.alignStream();
                pathId = this.reader.readInt64();
            }
            let byteStart;
            if (this.header.version >= 22) {
                byteStart = this.reader.readInt64();
            }
            else {
                byteStart = this.reader.readUInt32();
            }
            // @ts-ignore
            byteStart += this.header.dataOffset;
            const byteSize = this.reader.readUInt32();
            const typeId = this.reader.readInt32();
            let classId;
            let serializedType;
            if (this.header.version < 16) {
                classId = this.reader.readUInt16();
                serializedType = this.types.find((type) => type.classId === classId);
            }
            else {
                serializedType = this.types[typeId];
                classId = serializedType.classId;
            }
            if (this.header.version < 11) {
                this.reader.readUInt16(); // isDestroyed
            }
            if (this.header.version >= 11 && this.header.version < 17) {
                const scriptTypeIndex = this.reader.readInt16();
                if (serializedType) {
                    serializedType.scriptTypeIndex = scriptTypeIndex;
                }
            }
            if (this.header.version === 15 || this.header.version === 16) {
                this.reader.readInt8(); // stripped
            }
            objects.push({
                typeId,
                classId,
                pathId,
                byteStart,
                byteSize,
                serializedType,
            });
        }
        return objects;
    }
}
exports.SerializedFile = SerializedFile;
//# sourceMappingURL=SerializedFile.js.map